import tempfile
from pathlib import Path

import nbconvert
from fastapi.responses import FileResponse
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.nbconvert import Nbconvert


class _Nbconvert(Nbconvert):
    def __init__(
        self,
        app: App,
        auth: Auth,
    ) -> None:
        super().__init__(app, auth)

    async def get_nbconvert_formats(self):
        return {
            name: {"output_mimetype": nbconvert.exporters.get_exporter(name).output_mimetype}
            for name in nbconvert.exporters.get_export_names()
        }

    async def get_nbconvert_document(
        self,
        format: str,
        path: str,
        download: bool,
        user: User,
    ):
        exporter = nbconvert.exporters.get_exporter(format)
        if download:
            media_type = "application/octet-stream"
        else:
            media_type = exporter.output_mimetype
        tmp_dir = Path(tempfile.mkdtemp())
        tmp_path = tmp_dir / (Path(path).stem + exporter().file_extension)
        with open(tmp_path, "wt") as f:
            f.write(exporter().from_filename(path)[0])
        return FileResponse(tmp_path, media_type=media_type, filename=tmp_path.name)



from __future__ import annotations

import asyncio
import json
import os
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

import httpx
from fastapi import APIRouter, Cookie, Depends, HTTPException, Request, WebSocket, status
from fastapi.responses import RedirectResponse
from jupyterhub.services.auth import HubOAuth
from jupyterhub.utils import isoformat
from jupyverse_api import Router
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from .db import UserDB
from .models import JupyterHubUser


def auth_factory(
    app: App,
    db_session: AsyncSession,
    http_client: httpx.AsyncClient,
):
    class AuthJupyterHub(Auth, Router):
        def __init__(self) -> None:
            super().__init__(app)
            self.hub_auth = HubOAuth()
            self.db_lock = asyncio.Lock()
            self.activity_url = os.environ.get("JUPYTERHUB_ACTIVITY_URL")
            self.server_name = os.environ.get("JUPYTERHUB_SERVER_NAME")
            self.background_tasks = set()

            router = APIRouter()

            @router.get("/oauth_callback")
            async def get_oauth_callback(
                request: Request,
                code: str | None = None,
                state: str | None = None,
            ):
                if code is None:
                    raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

                cookie_state = request.cookies.get(self.hub_auth.state_cookie_name)
                if state is None or state != cookie_state:
                    raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

                token = self.hub_auth.token_for_code(code)
                hub_user = await self.hub_auth.user_for_token(token, use_cache=False, sync=False)
                async with self.db_lock:
                    db_session.add(
                        UserDB(
                            token=token,
                            name=hub_user["name"],
                            username=hub_user["name"],
                        ),
                    )
                    await db_session.commit()

                next_url = self.hub_auth.get_next_url(cookie_state)
                response = RedirectResponse(next_url)
                response.set_cookie(key="jupyverse_jupyterhub_token", value=token)
                return response

            @router.get("/api/me")
            async def get_api_me(
                request: Request,
                user: User = Depends(self.current_user()),
            ):
                checked_permissions: Dict[str, List[str]] = {}
                permissions = json.loads(
                    dict(request.query_params).get("permissions", "{}").replace("'", '"')
                )
                if permissions:
                    user_permissions: Dict[str, List[str]] = {}
                    for resource, actions in permissions.items():
                        user_resource_permissions = user_permissions.get(resource, [])
                        allowed = checked_permissions[resource] = []
                        for action in actions:
                            if action in user_resource_permissions:
                                allowed.append(action)

                keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
                identity = {k: getattr(user, k) for k in keys}
                return {
                    "identity": identity,
                    "permissions": checked_permissions,
                }

            self.include_router(router)

        def current_user(self, permissions: Optional[Dict[str, List[str]]] = None) -> Callable:
            async def _(
                request: Request,
                jupyverse_jupyterhub_token: Annotated[Union[str, None], Cookie()] = None,
            ):
                if jupyverse_jupyterhub_token is not None:
                    hub_user = await self.hub_auth.user_for_token(
                        jupyverse_jupyterhub_token, use_cache=False, sync=False
                    )
                    scopes = self.hub_auth.check_scopes(self.hub_auth.access_scopes, hub_user)
                    if not scopes:
                        raise HTTPException(
                            status_code=status.HTTP_403_FORBIDDEN,
                            detail=f"User {hub_user['name']} cannot access this server",
                        )

                    async with self.db_lock:
                        user_db = await db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                    user = JupyterHubUser.model_validate(user_db)
                    if self.activity_url:
                        headers = {
                            "Authorization": f"token {self.hub_auth.api_token}",
                            "Content-Type": "application/json",
                        }
                        last_activity = isoformat(datetime.utcnow())
                        task = asyncio.create_task(
                            http_client.post(
                                self.activity_url,
                                headers=headers,
                                json={
                                    "servers": {self.server_name: {"last_activity": last_activity}}
                                },
                            )
                        )
                        self.background_tasks.add(task)
                        task.add_done_callback(self.background_tasks.discard)
                    return user

                if permissions:
                    raise HTTPException(
                        status_code=status.HTTP_403_FORBIDDEN,
                    )

                state = self.hub_auth.generate_state(next_url=str(request.url))
                raise HTTPException(
                    status_code=status.HTTP_307_TEMPORARY_REDIRECT,
                    headers={
                        "Location": f"{self.hub_auth.login_url}&state={state}",
                        "Set-Cookie": f"{self.hub_auth.state_cookie_name}={state}",
                    },
                )

            return _

        async def update_user(
            self, jupyverse_jupyterhub_token: Annotated[Union[str, None], Cookie()] = None
        ) -> Callable:
            async def _(data: Dict[str, Any]) -> JupyterHubUser:
                if jupyverse_jupyterhub_token is not None:
                    async with self.db_lock:
                        user_db = await db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                        for k, v in data.items():
                            setattr(user_db, k, v)
                        await db_session.commit()
                    user = JupyterHubUser.model_validate(user_db)
                    return user

            return _

        def websocket_auth(
            self,
            permissions: Optional[Dict[str, List[str]]] = None,
        ) -> Callable[[], Tuple[Any, Dict[str, List[str]]]]:
            async def _(
                websocket: WebSocket,
            ) -> Optional[Tuple[WebSocket, Optional[Dict[str, List[str]]]]]:
                accept_websocket = False
                if "jupyverse_jupyterhub_token" in websocket._cookies:
                    jupyverse_jupyterhub_token = websocket._cookies["jupyverse_jupyterhub_token"]
                    async with self.db_lock:
                        user_db = await db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                    if user_db:
                        accept_websocket = True
                if accept_websocket:
                    return websocket, permissions
                else:
                    await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
                    return None

            return _

    return AuthJupyterHub()



INDEX_HTML = """\
<!doctype html><html lang="en"><head><meta charset="utf-8"><title>JupyterLab</title>
<meta name="viewport" content="width=device-width,initial-scale=1">
<script id="jupyter-config-data" type="application/json">PAGE_CONFIG</script>
VENDORS_NODE_MODULES
<script defer="defer" src="FULL_STATIC_URL/main.MAIN_ID.js?v=MAIN_ID"></script>
</head><body><script>/* Remove token from URL. */
  (function () {
    var location = window.location;
    var search = location.search;

    // If there is no query string, bail.
    if (search.length <= 1) {
      return;
    }

    // Rebuild the query string without the `token`.
    var query = '?' + search.slice(1).split('&')
      .filter(function (param) { return param.split('=')[0] !== 'token'; })
      .join('&');

    // Rebuild the URL with the new query string.
    var url = location.origin + location.pathname +
      (query !== '?' ? query : '') + location.hash;

    if (url === location.href) {
      return;
    }

    window.history.replaceState({ }, '', url);
  })();</script></body></html>
"""



import json
import struct
from typing import Any, Dict, List, Optional, Tuple

from zmq.asyncio import Socket

from ..kernel_driver.message import DELIM, deserialize, feed_identities, sign, unpack


def to_binary(msg: Dict[str, Any]) -> Optional[bytes]:
    if not msg["buffers"]:
        return None
    buffers = msg.pop("buffers")
    bmsg = json.dumps(msg).encode("utf8")
    buffers.insert(0, bmsg)
    n = len(buffers)
    offsets = [4 * (n + 1)]
    for b in buffers[:-1]:
        offsets.append(offsets[-1] + len(b))
    header = struct.pack("!" + "I" * (n + 1), n, *offsets)
    buffers.insert(0, header)
    return b"".join(buffers)


def from_binary(bmsg: bytes) -> Dict[str, Any]:
    n = struct.unpack("!i", bmsg[:4])[0]
    offsets = list(struct.unpack("!" + "I" * n, bmsg[4 : 4 * (n + 1)]))  # noqa
    offsets.append(None)
    buffers = []
    for start, stop in zip(offsets[:-1], offsets[1:]):
        buffers.append(bmsg[start:stop])
    msg = json.loads(buffers[0].decode("utf8"))
    msg["buffers"] = buffers[1:]
    return msg


async def send_raw_message(parts: List[bytes], sock: Socket, key: str) -> None:
    msg = parts[:4]
    buffers = parts[4:]
    to_send = [DELIM, sign(msg, key)] + msg + buffers
    await sock.send_multipart(to_send)


def deserialize_msg_from_ws_v1(ws_msg: bytes) -> Tuple[str, List[bytes]]:
    offset_number = int.from_bytes(ws_msg[:8], "little")
    offsets = [
        int.from_bytes(ws_msg[8 * (i + 1) : 8 * (i + 2)], "little")  # noqa
        for i in range(offset_number)
    ]
    channel = ws_msg[offsets[0] : offsets[1]].decode("utf-8")  # noqa
    msg_list = [ws_msg[offsets[i] : offsets[i + 1]] for i in range(1, offset_number - 1)]  # noqa
    return channel, msg_list


async def get_zmq_parts(socket: Socket) -> List[bytes]:
    parts = await socket.recv_multipart()
    idents, parts = feed_identities(parts)
    return parts


def get_msg_from_parts(
    parts: List[bytes], parent_header: Optional[Dict[str, Any]] = None
) -> Dict[str, Any]:
    return deserialize(parts, parent_header=parent_header)


def serialize_msg_to_ws_v1(msg_list: List[bytes], channel: str) -> List[bytes]:
    msg_list = msg_list[1:]
    channel_b = channel.encode("utf-8")
    offsets = []
    offsets.append(8 * (1 + 1 + len(msg_list) + 1))
    offsets.append(len(channel_b) + offsets[-1])
    for msg in msg_list:
        offsets.append(len(msg) + offsets[-1])
    offset_number = len(offsets).to_bytes(8, byteorder="little")
    offsets_b = [offset.to_bytes(8, byteorder="little") for offset in offsets]
    bin_msg = [offset_number] + offsets_b + [channel_b] + msg_list
    return bin_msg


def get_parent_header(parts: List[bytes]) -> Dict[str, Any]:
    return unpack(parts[2])



from abc import ABC, abstractmethod

from fastapi import APIRouter, Depends
from jupyverse_api import Router

from ..app import App
from ..auth import Auth, User


class Nbconvert(Router, ABC):
    def __init__(self, app: App, auth: Auth):
        super().__init__(app)

        router = APIRouter()

        @router.get("/api/nbconvert")
        async def get_nbconvert_formats():
            return await self.get_nbconvert_formats()

        @router.get("/nbconvert/{format}/{path}")
        async def get_nbconvert_document(
            format: str,
            path: str,
            download: bool,
            user: User = Depends(auth.current_user(permissions={"nbconvert": ["read"]})),
        ):
            return await self.get_nbconvert_document(format, path, download, user)

        self.include_router(router)

    @abstractmethod
    async def get_nbconvert_formats(self):
        ...

    @abstractmethod
    async def get_nbconvert_document(
        self,
        format: str,
        path: str,
        download: bool,
        user: User,
    ):
        ...



import asyncio
import json
from pathlib import Path

import pytest
import requests
import y_py as Y
from websockets import connect
from ypy_websocket import WebsocketProvider

prev_theme = {}
test_theme = {"raw": '{// jupyverse test\n"theme": "JupyterLab Dark"}'}


@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.parametrize("clear_users", (False,))
def test_settings_persistence_put(start_jupyverse):
    url = start_jupyverse
    # get previous theme
    response = requests.get(url + "/lab/api/settings/@jupyterlab/apputils-extension:themes")
    assert response.status_code == 200
    prev_theme["raw"] = json.loads(response.content)["raw"]
    # put new theme
    response = requests.put(
        url + "/lab/api/settings/@jupyterlab/apputils-extension:themes", data=json.dumps(test_theme)
    )
    assert response.status_code == 204


@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.parametrize("clear_users", (False,))
def test_settings_persistence_get(start_jupyverse):
    url = start_jupyverse
    # get new theme
    response = requests.get(
        url + "/lab/api/settings/@jupyterlab/apputils-extension:themes",
    )
    assert response.status_code == 200
    assert json.loads(response.content)["raw"] == test_theme["raw"]
    # put previous theme back
    response = requests.put(
        url + "/lab/api/settings/@jupyterlab/apputils-extension:themes",
        data=json.dumps(prev_theme),
    )
    assert response.status_code == 204


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.parametrize("clear_users", (False,))
async def test_rest_api(start_jupyverse):
    url = start_jupyverse
    ws_url = url.replace("http", "ws", 1)
    name = "notebook0.ipynb"
    path = (Path("tests") / "data" / name).as_posix()
    # create a session to launch a kernel
    response = requests.post(
        f"{url}/api/sessions",
        data=json.dumps(
            {
                "kernel": {"name": "python3"},
                "name": name,
                "path": path,
                "type": "notebook",
            }
        ),
    )
    r = response.json()
    kernel_id = r["kernel"]["id"]
    # get the room ID for the document
    response = requests.put(
        f"{url}/api/collaboration/session/{path}",
        data=json.dumps(
            {
                "format": "json",
                "type": "notebook",
            }
        ),
    )
    file_id = response.json()["fileId"]
    document_id = f"json:notebook:{file_id}"
    ydoc = Y.YDoc()
    async with connect(
        f"{ws_url}/api/collaboration/room/{document_id}"
    ) as websocket, WebsocketProvider(ydoc, websocket):
        # connect to the shared notebook document
        # wait for file to be loaded and Y model to be created in server and client
        await asyncio.sleep(0.5)
        # execute notebook
        for cell_idx in range(3):
            response = requests.post(
                f"{url}/api/kernels/{kernel_id}/execute",
                data=json.dumps(
                    {
                        "document_id": document_id,
                        "cell_idx": cell_idx,
                    }
                ),
            )
        # wait for Y model to be updated
        await asyncio.sleep(0.5)
        # retrieve cells
        cells = json.loads(ydoc.get_array("cells").to_json())
        assert cells[0]["outputs"] == [
            {
                "data": {"text/plain": ["3"]},
                "execution_count": 1.0,
                "metadata": {},
                "output_type": "execute_result",
            }
        ]
        assert cells[1]["outputs"] == [
            {"name": "stdout", "output_type": "stream", "text": ["Hello World!\n"]}
        ]
        assert cells[2]["outputs"] == [
            {
                "data": {"text/plain": ["7"]},
                "execution_count": 3.0,
                "metadata": {},
                "output_type": "execute_result",
            }
        ]



import os
from urllib.parse import unquote, urlparse

from jupyverse_api.cli import main


def launch():
    service_url = unquote(os.environ.get("JUPYTERHUB_SERVICE_URL"))
    url = urlparse(service_url)
    try:
        return main.callback(
            open_browser=True,
            host=url.hostname,
            port=url.port,
            set_=[
                f"frontend.base_url={url.path}",
                f"app.mount_path={url.path}",
            ],
            disable=[],
        )
    except Exception:
        return



from fastapi import Request, Response
from fastapi.responses import RedirectResponse


class RedirectException(Exception):
    def __init__(self, redirect_to: str):
        self.redirect_to = redirect_to


async def _redirect_exception_handler(request: Request, exc: RedirectException) -> Response:
    return RedirectResponse(url=exc.redirect_to)



from asphalt.core import Component, Context
from jupyverse_api.app import App

from .config import WebDAVConfig
from .routes import WebDAV


class WebDAVComponent(Component):
    def __init__(self, **kwargs):
        self.webdav_config = WebDAVConfig(**kwargs)

    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)

        webdav = WebDAV(app, self.webdav_config)
        ctx.add_resource(webdav)



from abc import ABC, abstractmethod

from fastapi import APIRouter, Depends, Request
from fastapi.responses import HTMLResponse
from jupyverse_api import Config, Router

from ..auth import Auth, User
from ..app import App


class JupyterLab(Router, ABC):
    def __init__(self, app: App, auth: Auth):
        super().__init__(app)

        router = APIRouter()

        @router.get("/lab")
        async def get_lab(
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_lab(user)

        @router.get("/lab/tree/{path:path}")
        async def load_workspace(
            path,
        ):
            return await self.load_workspace(path)

        @router.get("/lab/api/workspaces/{name}")
        async def get_workspace_data(user: User = Depends(auth.current_user())):
            return await self.get_workspace_data(user)

        @router.put(
            "/lab/api/workspaces/{name}",
            status_code=204,
        )
        async def set_workspace(
            request: Request,
            user: User = Depends(auth.current_user()),
            user_update=Depends(auth.update_user),
        ):
            return await self.set_workspace(request, user, user_update)

        @router.get("/lab/workspaces/{name}", response_class=HTMLResponse)
        async def get_workspace(
            name,
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_workspace(name, user)

        self.include_router(router)

    @abstractmethod
    async def get_lab(
        self,
        user: User,
    ):
        ...

    @abstractmethod
    async def load_workspace(
        self,
        path,
    ):
        ...

    @abstractmethod
    async def get_workspace_data(self, user: User):
        ...

    @abstractmethod
    async def set_workspace(
        self,
        request: Request,
        user: User,
        user_update,
    ):
        ...

    @abstractmethod
    async def get_workspace(
        self,
        name,
        user: User,
    ):
        ...


class JupyterLabConfig(Config):
    dev_mode: bool = False



import hashlib
import hmac
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, cast
from uuid import uuid4

from dateutil.parser import parse as dateutil_parse
from zmq.asyncio import Socket
from zmq.utils import jsonapi

protocol_version_info = (5, 3)
protocol_version = "%i.%i" % protocol_version_info

DELIM = b"<IDS|MSG>"


def feed_identities(msg_list: List[bytes]) -> Tuple[List[bytes], List[bytes]]:
    idx = msg_list.index(DELIM)
    return msg_list[:idx], msg_list[idx + 1 :]  # noqa


def str_to_date(obj: Dict[str, Any]) -> Dict[str, Any]:
    if "date" in obj:
        obj["date"] = dateutil_parse(obj["date"])
    return obj


def date_to_str(obj: Dict[str, Any]):
    if "date" in obj and not isinstance(obj["date"], str):
        obj["date"] = obj["date"].isoformat().replace("+00:00", "Z")
    return obj


def utcnow() -> datetime:
    return datetime.utcnow().replace(tzinfo=timezone.utc)


def create_message_header(msg_type: str, session_id: str, msg_id: str) -> Dict[str, Any]:
    if not session_id:
        session_id = msg_id = uuid4().hex
    else:
        msg_id = f"{session_id}_{msg_id}"
    header = {
        "date": utcnow().isoformat().replace("+00:00", "Z"),
        "msg_id": msg_id,
        "msg_type": msg_type,
        "session": session_id,
        "username": "",
        "version": protocol_version,
    }
    return header


def create_message(
    msg_type: str,
    content: Dict = {},
    session_id: str = "",
    msg_id: str = "",
) -> Dict[str, Any]:
    header = create_message_header(msg_type, session_id, msg_id)
    msg = {
        "header": header,
        "msg_id": header["msg_id"],
        "msg_type": header["msg_type"],
        "parent_header": {},
        "content": content,
        "metadata": {},
        "buffers": [],
    }
    return msg


def pack(obj: Dict[str, Any]) -> bytes:
    return jsonapi.dumps(obj)


def unpack(s: bytes) -> Dict[str, Any]:
    return cast(Dict[str, Any], jsonapi.loads(s))


def sign(msg_list: List[bytes], key: str) -> bytes:
    auth = hmac.new(key.encode("ascii"), digestmod=hashlib.sha256)
    h = auth.copy()
    for m in msg_list:
        h.update(m)
    return h.hexdigest().encode()


def serialize(msg: Dict[str, Any], key: str, change_date_to_str: bool = False) -> List[bytes]:
    _date_to_str = date_to_str if change_date_to_str else lambda x: x
    message = [
        pack(_date_to_str(msg["header"])),
        pack(_date_to_str(msg["parent_header"])),
        pack(_date_to_str(msg["metadata"])),
        pack(_date_to_str(msg.get("content", {}))),
    ]
    to_send = [DELIM, sign(message, key)] + message + msg.get("buffers", [])
    return to_send


def deserialize(
    msg_list: List[bytes],
    parent_header: Optional[Dict[str, Any]] = None,
    change_str_to_date: bool = False,
) -> Dict[str, Any]:
    _str_to_date = str_to_date if change_str_to_date else lambda x: x
    message: Dict[str, Any] = {}
    header = unpack(msg_list[1])
    message["header"] = _str_to_date(header)
    message["msg_id"] = header["msg_id"]
    message["msg_type"] = header["msg_type"]
    if parent_header:
        message["parent_header"] = parent_header
    else:
        message["parent_header"] = _str_to_date(unpack(msg_list[2]))
    message["metadata"] = unpack(msg_list[3])
    message["content"] = unpack(msg_list[4])
    message["buffers"] = [memoryview(b) for b in msg_list[5:]]
    return message


async def send_message(
    msg: Dict[str, Any], sock: Socket, key: str, change_date_to_str: bool = False
) -> None:
    await sock.send_multipart(serialize(msg, key, change_date_to_str=change_date_to_str), copy=True)


async def receive_message(
    sock: Socket, timeout: float = float("inf"), change_str_to_date: bool = False
) -> Optional[Dict[str, Any]]:
    timeout *= 1000  # in ms
    ready = await sock.poll(timeout)
    if ready:
        msg_list = await sock.recv_multipart()
        idents, msg_list = feed_identities(msg_list)
        return deserialize(msg_list, change_str_to_date=change_str_to_date)
    return None



from __future__ import annotations

from collections.abc import AsyncGenerator
from typing import Optional

from asphalt.core import Component, Context, context_teardown
from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.contents import Contents
from jupyverse_api.yjs import Yjs

from .routes import _Yjs


class YjsComponent(Component):
    @context_teardown
    async def start(
        self,
        ctx: Context,
    ) -> AsyncGenerator[None, Optional[BaseException]]:
        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore
        contents = await ctx.request_resource(Contents)  # type: ignore

        yjs = _Yjs(app, auth, contents)
        ctx.add_resource(yjs, types=Yjs)

        # start indexing in the background
        contents.file_id_manager

        yield

        yjs.room_manager.stop()
        contents.file_id_manager.stop_watching_files.set()
        await contents.file_id_manager.stopped_watching_files.wait()



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional

from fastapi import APIRouter, Depends, Request
from jupyverse_api import Router, Config

from ..app import App
from ..auth import Auth, User
from .models import Session


class Kernels(Router, ABC):
    def __init__(self, app: App, auth: Auth):
        super().__init__(app=app)

        router = APIRouter()

        @router.get("/api/status")
        async def get_status(
            user: User = Depends(auth.current_user(permissions={"status": ["read"]})),
        ):
            return await self.get_status(user)

        @router.get("/api/kernelspecs")
        async def get_kernelspecs(
            user: User = Depends(auth.current_user(permissions={"kernelspecs": ["read"]})),
        ):
            return await self.get_kernelspecs(user)

        @router.get("/kernelspecs/{kernel_name}/{file_name}")
        async def get_kernelspec(
            kernel_name,
            file_name,
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_kernelspec(kernel_name, file_name, user)

        @router.get("/api/kernels")
        async def get_kernels(
            user: User = Depends(auth.current_user(permissions={"kernels": ["read"]})),
        ):
            return await self.get_kernels(user)

        @router.delete("/api/sessions/{session_id}", status_code=204)
        async def delete_session(
            session_id: str,
            user: User = Depends(auth.current_user(permissions={"sessions": ["write"]})),
        ):
            return await self.delete_session(session_id, user)

        @router.patch("/api/sessions/{session_id}")
        async def rename_session(
            request: Request,
            user: User = Depends(auth.current_user(permissions={"sessions": ["write"]})),
        ) -> Session:
            return await self.rename_session(request, user)

        @router.get("/api/sessions")
        async def get_sessions(
            user: User = Depends(auth.current_user(permissions={"sessions": ["read"]})),
        ) -> List[Session]:
            return await self.get_sessions(user)

        @router.post(
            "/api/sessions",
            status_code=201,
            response_model=Session,
        )
        async def create_session(
            request: Request,
            user: User = Depends(auth.current_user(permissions={"sessions": ["write"]})),
        ) -> Session:
            return await self.create_session(request, user)

        @router.post("/api/kernels/{kernel_id}/interrupt")
        async def interrupt_kernel(
            kernel_id,
            user: User = Depends(auth.current_user(permissions={"kernels": ["write"]})),
        ):
            return await self.interrupt_kernel(kernel_id, user)

        @router.post("/api/kernels/{kernel_id}/restart")
        async def restart_kernel(
            kernel_id,
            user: User = Depends(auth.current_user(permissions={"kernels": ["write"]})),
        ):
            return await self.restart_kernel(kernel_id, user)

        @router.post("/api/kernels/{kernel_id}/execute")
        async def execute_cell(
            request: Request,
            kernel_id,
            user: User = Depends(auth.current_user(permissions={"kernels": ["write"]})),
        ):
            return await self.execute_cell(request, kernel_id, user)

        @router.get("/api/kernels/{kernel_id}")
        async def get_kernel(
            kernel_id,
            user: User = Depends(auth.current_user(permissions={"kernels": ["read"]})),
        ):
            return await self.get_kernel(kernel_id, user)

        @router.delete("/api/kernels/{kernel_id}", status_code=204)
        async def shutdown_kernel(
            kernel_id,
            user: User = Depends(auth.current_user(permissions={"kernels": ["write"]})),
        ):
            return await self.shutdown_kernel(kernel_id, user)

        @router.websocket("/api/kernels/{kernel_id}/channels")
        async def kernel_channels(
            kernel_id,
            session_id,
            websocket_permissions=Depends(
                auth.websocket_auth(permissions={"kernels": ["execute"]})
            ),
        ):
            return await self.kernel_channels(kernel_id, session_id, websocket_permissions)

        self.include_router(router)

    @abstractmethod
    async def watch_connection_files(self, path: Path) -> None:
        ...

    @abstractmethod
    async def get_status(
        self,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_kernelspecs(
        self,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_kernelspec(
        self,
        kernel_name,
        file_name,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_kernels(
        self,
        user: User,
    ):
        ...

    @abstractmethod
    async def delete_session(
        self,
        session_id: str,
        user: User,
    ):
        ...

    @abstractmethod
    async def rename_session(
        self,
        request: Request,
        user: User,
    ) -> Session:
        ...

    @abstractmethod
    async def get_sessions(
        self,
        user: User,
    ) -> List[Session]:
        ...

    @abstractmethod
    async def create_session(
        self,
        request: Request,
        user: User,
    ) -> Session:
        ...

    @abstractmethod
    async def interrupt_kernel(
        self,
        kernel_id,
        user: User,
    ):
        ...

    @abstractmethod
    async def restart_kernel(
        self,
        kernel_id,
        user: User,
    ):
        ...

    @abstractmethod
    async def execute_cell(
        self,
        request: Request,
        kernel_id,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_kernel(
        self,
        kernel_id,
        user: User,
    ):
        ...

    @abstractmethod
    async def shutdown_kernel(
        self,
        kernel_id,
        user: User,
    ):
        ...

    @abstractmethod
    async def kernel_channels(
        self,
        kernel_id,
        session_id,
        websocket_permissions,
    ):
        ...


class KernelsConfig(Config):
    default_kernel: str = "python3"
    connection_path: Optional[str] = None
    require_yjs: bool = False



import asyncio
import json
import os
import socket
import sys
import tempfile
import uuid
from typing import Dict, Optional, Tuple, Union

import zmq
import zmq.asyncio
from zmq.asyncio import Socket

channel_socket_types = {
    "hb": zmq.REQ,
    "shell": zmq.DEALER,
    "iopub": zmq.SUB,
    "stdin": zmq.DEALER,
    "control": zmq.DEALER,
}

context = zmq.asyncio.Context()

cfg_t = Dict[str, Union[str, int]]


def get_port(ip: str) -> int:
    sock = socket.socket()
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, b"\0" * 8)
    sock.bind((ip, 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def write_connection_file(
    fname: str = "",
    ip: str = "",
    transport: str = "tcp",
    signature_scheme: str = "hmac-sha256",
    kernel_name: str = "",
) -> Tuple[str, cfg_t]:
    ip = ip or "127.0.0.1"

    if not fname:
        fd, fname = tempfile.mkstemp(suffix=".json")
        os.close(fd)
    f = open(fname, "wt")

    channels = ["shell", "iopub", "stdin", "control", "hb"]

    cfg: cfg_t = {f"{c}_port": get_port(ip) for c in channels}

    cfg["ip"] = ip
    cfg["key"] = uuid.uuid4().hex
    cfg["transport"] = transport
    cfg["signature_scheme"] = signature_scheme
    cfg["kernel_name"] = kernel_name

    f.write(json.dumps(cfg, indent=2))
    f.close()

    return fname, cfg


def read_connection_file(fname: str) -> cfg_t:
    with open(fname, "rt") as f:
        cfg: cfg_t = json.load(f)

    return cfg


async def launch_kernel(
    kernelspec_path: str, connection_file_path: str, kernel_cwd: str, capture_output: bool
) -> asyncio.subprocess.Process:
    with open(kernelspec_path) as f:
        kernelspec = json.load(f)
    cmd = [s.format(connection_file=connection_file_path) for s in kernelspec["argv"]]
    if cmd and cmd[0] in {
        "python",
        "python%i" % sys.version_info[0],
        "python%i.%i" % sys.version_info[:2],
    }:
        cmd[0] = sys.executable
    if kernel_cwd:
        prev_dir = os.getcwd()
        os.chdir(kernel_cwd)
    if capture_output:
        p = await asyncio.create_subprocess_exec(
            *cmd, stdout=asyncio.subprocess.DEVNULL, stderr=asyncio.subprocess.STDOUT
        )
    else:
        p = await asyncio.create_subprocess_exec(*cmd)
    if kernel_cwd:
        os.chdir(prev_dir)
    return p


def create_socket(channel: str, cfg: cfg_t, identity: Optional[bytes] = None) -> Socket:
    ip = cfg["ip"]
    port = cfg[f"{channel}_port"]
    url = f"tcp://{ip}:{port}"
    socket_type = channel_socket_types[channel]
    sock = context.socket(socket_type)
    sock.linger = 1000  # set linger to 1s to prevent hangs at exit
    if identity:
        sock.identity = identity
    sock.connect(url)
    return sock


def connect_channel(channel_name: str, cfg: cfg_t, identity: Optional[bytes] = None) -> Socket:
    sock = create_socket(channel_name, cfg, identity)
    if channel_name == "iopub":
        sock.setsockopt(zmq.SUBSCRIBE, b"")
    return sock



from __future__ import annotations

import logging
from functools import partial
from pathlib import Path

from asgi_middleware_static_file import ASGIMiddlewareStaticFile  # type: ignore
from asgi_webdav.middleware.cors import ASGIMiddlewareCORS  # type: ignore
from asgi_webdav import __name__ as app_name  # type: ignore
from asgi_webdav import __version__  # type: ignore

try:
    from asgi_webdav.config import (  # type: ignore
        get_config,
        init_config_from_file,
        init_config_from_obj,
    )
    from asgi_webdav.constants import DAV_METHODS, AppEntryParameters  # type: ignore
    from asgi_webdav.server import Server  # type: ignore

    asgi_webdav_installed = True
except BaseException:
    asgi_webdav_installed = False
from jupyverse_api.app import App

from .config import WebDAVConfig


logger = logging.getLogger("webdav")


class WebDAVApp:
    def __init__(self, app, webdav_app):
        self._app = app
        self._webdav_app = webdav_app

    async def __call__(self, scope, receive, send):
        if scope.get("method") in DAV_METHODS and scope.get("path").startswith("/webdav"):
            return await self._webdav_app(scope, receive, send)
        return await self._app(scope, receive, send)


class WebDAV:
    def __init__(self, app: App, webdav_config: WebDAVConfig):
        if not asgi_webdav_installed:
            return

        for account in webdav_config.account_mapping:
            logger.info(f"WebDAV user {account.username} has password {account.password}")
        webdav_conf = webdav_config.dict()
        init_config_from_obj(webdav_conf)
        webdav_aep = AppEntryParameters()
        webdav_app = get_asgi_app(aep=webdav_aep, config_obj=webdav_conf)
        app.add_middleware(partial(WebDAVApp, webdav_app=webdav_app))


# this is to get rid of asgi-webdav's logging configuration, see:
# https://github.com/rexzhang/asgi-webdav/blob/53735fa67030e1db0d610deb58d2ebfedbdd7c3b/asgi_webdav/server.py#L99
def get_asgi_app(aep: AppEntryParameters, config_obj: dict | None = None):
    """create ASGI app"""
    # init config
    if aep.config_file is not None:
        init_config_from_file(aep.config_file)
    if config_obj is not None:
        init_config_from_obj(config_obj)

    config = get_config()
    config.update_from_app_args_and_env_and_default_value(aep=aep)

    # create ASGI app
    app = Server(config)

    # route /_/static
    app = ASGIMiddlewareStaticFile(
        app=app,
        static_url="_/static",
        static_root_paths=[Path(__file__).parent.joinpath("static")],
    )

    # CORS
    if config.cors.enable:
        app = ASGIMiddlewareCORS(
            app=app,
            allow_url_regex=config.cors.allow_url_regex,
            allow_origins=config.cors.allow_origins,
            allow_origin_regex=config.cors.allow_origin_regex,
            allow_methods=config.cors.allow_methods,
            allow_headers=config.cors.allow_headers,
            allow_credentials=config.cors.allow_credentials,
            expose_headers=config.cors.expose_headers,
            preflight_max_age=config.cors.preflight_max_age,
        )

    # config sentry
    if config.sentry_dsn:
        try:
            import sentry_sdk  # type: ignore
            from sentry_sdk.integrations.asgi import SentryAsgiMiddleware  # type: ignore

            sentry_sdk.init(
                dsn=config.sentry_dsn,
                release=f"{app_name}@{__version__}",
            )
            app = SentryAsgiMiddleware(app)

        except ImportError as e:
            logger.warning(e)

    return app



import asyncio
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional, Union

from fastapi import APIRouter, Depends, Request, Response
from jupyverse_api import Router

from .models import Checkpoint, Content, SaveContent
from ..auth import Auth, User
from ..app import App


class FileIdManager(ABC):
    stop_watching_files: asyncio.Event
    stopped_watching_files: asyncio.Event

    @abstractmethod
    async def get_path(self, file_id: str) -> str:
        ...

    @abstractmethod
    async def get_id(self, file_path: str) -> str:
        ...

    def watch(self, path: str):
        ...

    def unwatch(self, path: str, watcher):
        ...


class Contents(Router, ABC):
    def __init__(self, app: App, auth: Auth):
        super().__init__(app=app)

        router = APIRouter()

        @router.post(
            "/api/contents/{path:path}/checkpoints",
            status_code=201,
        )
        async def create_checkpoint(
            path, user: User = Depends(auth.current_user(permissions={"contents": ["write"]}))
        ) -> Checkpoint:
            return await self.create_checkpoint(path, user)

        @router.post(
            "/api/contents{path:path}",
            status_code=201,
        )
        async def create_content(
            path: Optional[str],
            request: Request,
            user: User = Depends(auth.current_user(permissions={"contents": ["write"]})),
        ) -> Content:
            return await self.create_content(path, request, user)

        @router.get("/api/contents")
        async def get_root_content(
            content: int,
            user: User = Depends(auth.current_user(permissions={"contents": ["read"]})),
        ) -> Content:
            return await self.get_root_content(content, user)

        @router.get("/api/contents/{path:path}/checkpoints")
        async def get_checkpoint(
            path, user: User = Depends(auth.current_user(permissions={"contents": ["read"]}))
        ) -> List[Checkpoint]:
            return await self.get_checkpoint(path, user)

        @router.get("/api/contents/{path:path}")
        async def get_content(
            path: str,
            content: int = 0,
            user: User = Depends(auth.current_user(permissions={"contents": ["read"]})),
        ) -> Content:
            return await self.get_content(path, content, user)

        @router.put("/api/contents/{path:path}")
        async def save_content(
            path,
            request: Request,
            response: Response,
            user: User = Depends(auth.current_user(permissions={"contents": ["write"]})),
        ) -> Content:
            return await self.save_content(path, request, response, user)

        @router.delete(
            "/api/contents/{path:path}",
            status_code=204,
        )
        async def delete_content(
            path,
            user: User = Depends(auth.current_user(permissions={"contents": ["write"]})),
        ):
            return await self.delete_content(path, user)

        @router.patch("/api/contents/{path:path}")
        async def rename_content(
            path,
            request: Request,
            user: User = Depends(auth.current_user(permissions={"contents": ["write"]})),
        ) -> Content:
            return await self.rename_content(path, request, user)

        self.include_router(router)

    @property
    @abstractmethod
    def file_id_manager(self) -> FileIdManager:
        ...

    @abstractmethod
    async def read_content(
        self, path: Union[str, Path], get_content: bool, file_format: Optional[str] = None
    ) -> Content:
        ...

    @abstractmethod
    async def write_content(self, content: Union[SaveContent, Dict]) -> None:
        ...

    @abstractmethod
    async def create_checkpoint(
        self,
        path,
        user: User,
    ) -> Checkpoint:
        ...

    @abstractmethod
    async def create_content(
        self,
        path: Optional[str],
        request: Request,
        user: User,
    ) -> Content:
        ...

    @abstractmethod
    async def get_root_content(
        self,
        content: int,
        user: User,
    ) -> Content:
        return await self.get_root_content(content, user)

    @abstractmethod
    async def get_checkpoint(
        self,
        path,
        user: User,
    ) -> List[Checkpoint]:
        ...

    @abstractmethod
    async def get_content(
        self,
        path: str,
        content: int,
        user: User,
    ) -> Content:
        ...

    @abstractmethod
    async def save_content(
        self,
        path,
        request: Request,
        response: Response,
        user: User,
    ) -> Content:
        ...

    @abstractmethod
    async def delete_content(
        self,
        path,
        user: User,
    ):
        ...

    @abstractmethod
    async def rename_content(
        self,
        path,
        request: Request,
        user: User,
    ) -> Content:
        ...



import pytest
from asphalt.core import Context
from fastapi import APIRouter
from httpx import AsyncClient
from jupyverse_api import Router
from jupyverse_api.app import App
from jupyverse_api.main import JupyverseComponent

from utils import configure


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "mount_path",
    (
        None,
        "/foo",
    ),
)
async def test_mount_path(mount_path, unused_tcp_port):
    components = configure({"app": {"type": "app"}}, {"app": {"mount_path": mount_path}})

    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        app = await ctx.request_resource(App)
        router = APIRouter()

        @router.get("/")
        async def get():
            pass

        Router(app).include_router(router)

        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}")
        expected = 200 if mount_path is None else 404
        assert response.status_code == expected

        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/bar")
        assert response.status_code == 404

        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/foo")
        expected = 404 if mount_path is None else 200



from asphalt.core import Component, Context
from jupyverse_api.auth import Auth
from jupyverse_api.app import App
from jupyverse_api.nbconvert import Nbconvert

from .routes import _Nbconvert


class NbconvertComponent(Component):
    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore

        nbconvert = _Nbconvert(app, auth)
        ctx.add_resource(nbconvert, types=Nbconvert)



import logging
import uuid
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, Tuple, cast

import httpx
from fastapi import Depends, HTTPException, Response, WebSocket, status
from fastapi_users import (
    BaseUserManager,
    FastAPIUsers,
    UUIDIDMixin,
    models,
)
from fastapi_users.authentication import (
    AuthenticationBackend,
    CookieTransport,
    JWTStrategy,
)
from fastapi_users.authentication.strategy.base import Strategy
from fastapi_users.authentication.transport.base import Transport
from fastapi_users.db import SQLAlchemyUserDatabase
from httpx_oauth.clients.github import GitHubOAuth2
from jupyverse_api.exceptions import RedirectException
from jupyverse_api.frontend import FrontendConfig
from starlette.requests import Request

from .config import _AuthConfig
from .db import User
from .models import UserCreate, UserRead


logger = logging.getLogger("auth")


@dataclass
class Res:
    cookie_authentication: Any
    current_user: Any
    update_user: Any
    fapi_users: Any
    get_user_manager: Any
    github_authentication: Any
    github_cookie_authentication: Any
    websocket_auth: Any


def get_backend(auth_config: _AuthConfig, frontend_config: FrontendConfig, db) -> Res:
    class NoAuthScheme:
        def __call__(self):
            return "noauth"

    class NoAuthTransport(Transport):
        scheme = NoAuthScheme()  # type: ignore

        async def get_login_response(self, token: str) -> Response:
            return Response()

        async def get_logout_response(self) -> Response:
            return Response()

        @staticmethod
        def get_openapi_login_responses_success():
            pass

        @staticmethod
        def get_openapi_logout_responses_success():
            pass

    class NoAuthStrategy(Strategy, Generic[models.UP, models.ID]):
        async def read_token(
            self, token: Optional[str], user_manager: BaseUserManager[models.UP, models.ID]
        ) -> Optional[models.UP]:
            active_user = await user_manager.user_db.get_by_email(auth_config.global_email)
            return active_user

        async def write_token(self, user: models.UP):
            pass

        async def destroy_token(self, token: str, user: models.UP):
            pass

    class GitHubTransport(CookieTransport):
        async def get_login_response(self, token: str) -> Response:
            response = await super().get_login_response(token)
            response.status_code = status.HTTP_302_FOUND
            response.headers["Location"] = "/lab"
            return response

    def get_noauth_strategy() -> NoAuthStrategy:
        return NoAuthStrategy()

    def get_jwt_strategy() -> JWTStrategy:
        return JWTStrategy(secret=db.secret, lifetime_seconds=None)

    noauth_authentication = AuthenticationBackend(
        name="noauth",
        transport=NoAuthTransport(),
        get_strategy=get_noauth_strategy,
    )

    cookie_authentication = AuthenticationBackend(
        name="cookie",
        transport=CookieTransport(cookie_secure=auth_config.cookie_secure),
        get_strategy=get_jwt_strategy,
    )

    github_cookie_authentication = AuthenticationBackend(
        name="github",
        transport=GitHubTransport(),
        get_strategy=get_jwt_strategy,
    )

    github_authentication = GitHubOAuth2(auth_config.client_id, auth_config.client_secret)

    class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
        async def on_after_register(self, user: User, request: Optional[Request] = None):
            for oauth_account in await user.awaitable_attrs.oauth_accounts:
                if oauth_account.oauth_name == "github":
                    async with httpx.AsyncClient() as client:
                        r = (
                            await client.get(
                                f"https://api.github.com/user/{oauth_account.account_id}"
                            )
                        ).json()

                    await self.user_db.update(
                        user,
                        dict(
                            anonymous=False,
                            username=r["login"],
                            color=None,
                            avatar_url=r["avatar_url"],
                            is_active=True,
                        ),
                    )

    async def get_user_manager(user_db: SQLAlchemyUserDatabase = Depends(db.get_user_db)):
        yield UserManager(user_db)

    def get_enabled_backends():
        if auth_config.mode == "noauth" and not frontend_config.collaborative:
            res = [noauth_authentication, github_cookie_authentication]
        else:
            res = [cookie_authentication, github_cookie_authentication]
        return res

    fapi_users = FastAPIUsers[User, uuid.UUID](
        get_user_manager,
        [
            noauth_authentication,
            cookie_authentication,
            github_cookie_authentication,
        ],
    )

    async def create_guest(user_manager):
        # workspace and settings are copied from global user
        # but this is a new user
        global_user = await user_manager.get_by_email(auth_config.global_email)
        user_id = str(uuid.uuid4())
        guest = dict(
            anonymous=True,
            email=f"{user_id}@jupyter.com",
            username=f"{user_id}@jupyter.com",
            password="",
            workspace=global_user.workspace,
            settings=global_user.settings,
            permissions={},
        )
        return await user_manager.create(UserCreate(**guest))

    def current_user(permissions: Optional[Dict[str, List[str]]] = None):
        async def _(
            response: Response,
            token: Optional[str] = None,
            user: Optional[User] = Depends(
                fapi_users.current_user(optional=True, get_enabled_backends=get_enabled_backends)
            ),
            user_manager: BaseUserManager[User, models.ID] = Depends(get_user_manager),
        ):
            if auth_config.mode == "user":
                # "user" authentication: check authorization
                if user and permissions:
                    for resource, actions in permissions.items():
                        user_actions_for_resource = user.permissions.get(resource, [])
                        if not all([a in user_actions_for_resource for a in actions]):
                            user = None
                            break
            else:
                # "noauth" or "token" authentication
                if frontend_config.collaborative:
                    if not user and auth_config.mode == "noauth":
                        user = await create_guest(user_manager)
                        token = await get_jwt_strategy().write_token(user)
                        cookie_transport = cast(CookieTransport, cookie_authentication.transport)
                        cookie_transport._set_login_cookie(response, token)

                    elif not user and auth_config.mode == "token":
                        global_user = await user_manager.get_by_email(auth_config.global_email)
                        if global_user and global_user.username == token:
                            user = await create_guest(user_manager)
                            token = await get_jwt_strategy().write_token(user)
                            cookie_transport = cast(
                                CookieTransport, cookie_authentication.transport
                            )
                            cookie_transport._set_login_cookie(response, token)
                else:
                    if auth_config.mode == "token":
                        global_user = await user_manager.get_by_email(auth_config.global_email)
                        if global_user and global_user.username == token:
                            user = global_user
                            token = await get_jwt_strategy().write_token(user)
                            cookie_transport = cast(
                                CookieTransport, cookie_authentication.transport
                            )
                            cookie_transport._set_login_cookie(response, token)

            if user:
                return user

            elif auth_config.login_url:
                raise RedirectException(auth_config.login_url)

            else:
                raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

        return _

    def websocket_auth(permissions: Optional[Dict[str, List[str]]] = None):
        """
        A function returning a dependency for the WebSocket connection.

        :param permissions: the permissions the user should be granted access to. The user should
        have access to at least one of them for the WebSocket to be opened.
        :returns: a dependency for the WebSocket connection. The dependency returns a tuple
        consisting of the websocket and the checked user permissions if the websocket is accepted,
        None otherwise.
        """

        async def _(
            websocket: WebSocket,
            user_manager: BaseUserManager[models.UP, models.ID] = Depends(get_user_manager),
        ) -> Optional[Tuple[WebSocket, Optional[Dict[str, List[str]]]]]:
            accept_websocket = False
            checked_permissions: Optional[Dict[str, List[str]]] = None
            if auth_config.mode == "noauth":
                accept_websocket = True
            elif "fastapiusersauth" in websocket._cookies:
                token = websocket._cookies["fastapiusersauth"]
                user = await get_jwt_strategy().read_token(token, user_manager)
                if user:
                    if auth_config.mode == "user":
                        # "user" authentication: check authorization
                        if permissions is None:
                            accept_websocket = True
                        else:
                            checked_permissions = {}
                            for resource, actions in permissions.items():
                                user_actions_for_resource = user.permissions.get(resource)
                                if user_actions_for_resource is None:
                                    continue
                                allowed = checked_permissions[resource] = []
                                for action in actions:
                                    if action in user_actions_for_resource:
                                        allowed.append(action)
                                        accept_websocket = True
                    else:
                        accept_websocket = True
            if accept_websocket:
                return websocket, checked_permissions
            else:
                await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
                return None

        return _

    async def update_user(
        user: UserRead = Depends(current_user()),
        user_db: SQLAlchemyUserDatabase = Depends(db.get_user_db),
    ):
        async def _(data: Dict[str, Any]) -> UserRead:
            await user_db.update(user, data)
            return user

        return _

    return Res(
        cookie_authentication=cookie_authentication,
        current_user=current_user,
        update_user=update_user,
        fapi_users=fapi_users,
        get_user_manager=get_user_manager,
        github_authentication=github_authentication,
        github_cookie_authentication=github_cookie_authentication,
        websocket_auth=websocket_auth,
    )



from asphalt.core import Component, Context
from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.jupyterlab import JupyterLabConfig
from jupyverse_api.lab import Lab

from .routes import _Lab


class LabComponent(Component):
    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore
        frontend_config = await ctx.request_resource(FrontendConfig)
        jupyterlab_config = ctx.get_resource(JupyterLabConfig)

        lab = _Lab(app, auth, frontend_config, jupyterlab_config)
        ctx.add_resource(lab, types=Lab)



from jupyverse_api.auth import AuthConfig
from pydantic import Field


class _AuthFiefConfig(AuthConfig):
    base_url: str = Field(description="Base URL of Fief tenant")
    callback_url: str = Field(description="URL of the callback route", default="")
    client_id: str = Field(description="ID of Fief client")
    client_secret: str = Field(description="Secret of Fief client")
    admin_api_key: str = Field(description="Admin API key", default="")
    oauth_provider_id: str = Field(description="OAuth provider ID", default="")
